import enum
import ipaddress
import math


class BitStreamReader:
    """Class for parsing streams bitwise"""

    def __init__(self, value):
        self._start = 0
        self._value = value

    def advance(self, length):
        """Advance the stream length bits"""
        self._start += length

    def bit(self, advance: bool = True) -> int:
        """Return the next bit in the stream"""
        return self.bits(1, advance)

    def bits(self, size: int = 8, advance=True) -> int:
        """Return size number of bits in the stream"""
        startBit = self._start % 8
        byte = (self.peekByte() << startBit) & 0xFF  # Mask off top part of byte
        byte = (byte >> 8 - size) & 0xFF  # Shift down and mask
        if advance:
            self.advance(size)
        return byte

    def byte(self, advance: bool = True) -> int:
        """Return one byte from the stream"""
        return int.from_bytes(self.value(1, advance), "little", signed=False)

    def bytesLeft(self) -> int:
        """Return the number of bytes remaining from the stream"""
        startByte = int(self._start / 8)
        return len(self._value) - startByte

    def peekByte(self) -> int:
        """Return the next byte from the stream without advancing the stream"""
        return self.byte(advance=False)

    def peekValue(self, size: int) -> bytes:
        """Return the next value from the stream without advancing the stream"""
        return self.value(size, advance=False)

    def remaining(self, advance: bool = True) -> bytes:
        """Return all the remaining bytes in the stream"""
        startByte = int(self._start / 8)
        if advance:
            self.advance(len(self._value) * 8)
        return self._value[startByte:]

    def value(self, size: int, advance: bool = True) -> bytes:
        """Return the next size number of bytes from the stream"""
        startByte = int(self._start / 8)
        if advance:
            self.advance(size * 8)
        if startByte + size > len(self._value):
            raise EOFError("Tried to read past end of data")
        return self._value[startByte : startByte + size]


class BitStreamWriter(bytearray):
    """Class for wringing a butearray bitwise"""

    def __init__(self):
        super().__init__()
        self._start = 0

    def addBits(self, value, size):
        """Add size number of bits to the stream"""
        byte = 0
        if self._start == 0:
            # Start new byte
            self.append(0)
        else:
            # Extract the last one
            byte = self[len(self) - 1]
        self[len(self) - 1] = byte | (value << 8 - self._start - size)
        self._start = (self._start + size) % 8

    def addBytes(self, value, size, signed, endian="big"):
        """Add size number of bytes to the stream"""
        newVal = value.to_bytes(size, endian, signed=signed)
        self.extend(newVal)
        self._start = 0


class str_t(str):  # pylint: disable=invalid-name
    """Unicode string"""

    def __getstate__(self):
        return str(self)

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize unicode string"""
        length = stream.byte()
        return stream.value(length).decode("utf-8")


class int_t(int):  # pylint: disable=invalid-name
    """Base class for any int like type"""

    endian = "big"
    signed = True
    size = 0

    def __getstate__(self):
        return int(self)

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize unsigned value from stream"""
        return cls.from_bytes(stream.value(cls.size), cls.endian, signed=cls.signed)

    def serialize(self, stream: BitStreamWriter):
        """Serialize into stream"""
        stream.addBytes(self, self.size, self.signed, self.endian)


class uint_t(int_t):  # pylint: disable=invalid-name
    """Base class for any unsigned int like type"""

    signed = False
    size = 1

    def __repr__(self):
        return "0x{0:0{1}X} ({0})".format(self, self.size)


class int24_t(int_t):  # pylint: disable=invalid-name
    """Signed 24 bits value"""

    size = 3


class uint3_t(int):  # pylint: disable=invalid-name
    """Type representing 3 bits value"""

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize bits from stream"""
        return stream.bits(3)

    def serialize(self, stream: BitStreamWriter):
        """Serialize bits into stream"""
        stream.addBits(self, 3)


class uint4_t(int):  # pylint: disable=invalid-name
    """Type representing 4 bits value"""

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize bits from stream"""
        return stream.bits(4)

    def serialize(self, stream: BitStreamWriter):
        """Serialize bits into stream"""
        stream.addBits(self, 4)


class uint5_t(int):  # pylint: disable=invalid-name
    """Type representing 5 bits value"""

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize bits from stream"""
        return stream.bits(5)

    def serialize(self, stream: BitStreamWriter):
        """Serialize bits into stream"""
        stream.addBits(self, 5)


class uint7_t(int):  # pylint: disable=invalid-name
    """Type representing 7 bits value"""

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize bits from stream"""
        return stream.bits(7)

    def serialize(self, stream: BitStreamWriter):
        """Serialize bits into stream"""
        stream.addBits(self, 7)


class uint8_t(uint_t):  # pylint: disable=invalid-name
    """Unsigned byte"""

    size = 1


class uint16_t(uint_t):  # pylint: disable=invalid-name
    """Unsigned word"""

    size = 2


class uint32_t(uint_t):  # pylint: disable=invalid-name
    """Unsigned 32 bits value"""

    size = 4


class BitsBase:
    """Base type for bit values"""

    sizeBits = 1

    def __init__(self, value: int):
        self._value = int(value)

    def __eq__(self, other):
        return self._value.__eq__(other)

    def __int__(self):
        return self._value

    def __repr__(self):
        return "bits_t({0:0{1}b})".format(self._value, self.sizeBits)

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize bits from stream"""
        return stream.bits(cls.sizeBits)

    def serialize(self, stream: BitStreamWriter):
        """Serialize bits into stream"""
        stream.addBits(self._value, self.sizeBits)


class bytes_t(bytes):  # pylint: disable=invalid-name
    """Variable size bytes"""

    default = b""

    def serialize(self, stream: BitStreamWriter):
        """Serialize into stream"""
        stream.extend(self)

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize bytes from stream"""
        return stream.remaining()


class flag_t(BitsBase):  # pylint: disable=invalid-name
    """Type represeting one bit"""

    def __init__(self, value: int):
        super().__init__(int(value) & 1)

    def __bool__(self):
        return bool(self._value)

    def __repr__(self):
        return "flag_t({})".format(bool(self))


def bits_t(size):  # pylint: disable=invalid-name
    """Return the type for size number of bits"""
    # pylint: disable=redefined-outer-name
    class bits_t(BitsBase):  # pylint: disable=invalid-name
        """Type represting arbitrary bits"""

        sizeBits = size

    return bits_t


def enum_t(enumType, baseType):  # pylint: disable=invalid-name
    """Return a new enum type based on the specified type"""
    # pylint: disable=redefined-outer-name
    class enum_t(baseType):  # pylint: disable=invalid-name
        """Type for representing an enum value"""

        def __repr__(self):
            value = int(self)
            try:
                if issubclass(enumType, enum.IntFlag):
                    return "{0} ({1:b})".format(str(enumType(value)), value)
                return "{} (0x{:X})".format(enumType(value).name, value)
            except Exception:
                # Enum not defined
                return "UNKNOWN (0x{:X})".format(value)

    return enum_t


class float_t(float):  # pylint: disable=invalid-name
    """Type for representing signed float values."""

    def __new__(cls, value=0, *_args):  # pylint: disable=keyword-arg-before-vararg
        # Override __new__ to allow *args
        return float.__new__(cls, value)

    def __init__(self, _value=0, size=1, scale=0):
        super().__init__()
        self._size = size
        self._scale = scale

    @property
    def scale(self) -> int:
        """The scale this value represents"""
        return self._scale

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize float value from stream"""
        precision = stream.bits(3)
        scale = stream.bits(2)
        size = stream.bits(3)
        value = int.from_bytes(stream.value(size), "big", signed=True) / math.pow(
            10, precision
        )
        return (value, size, scale)


def reserved_t(size):  # pylint: disable=invalid-name
    """Return the type for bits that are reserved and must not be used"""
    # pylint: disable=redefined-outer-name
    class reserved_t(BitsBase):  # pylint: disable=invalid-name
        """Type for when the bits are reserved and must not be used"""

        default = 0
        sizeBits = size

    return reserved_t


class dsk_t:  # pylint: disable=invalid-name
    """Type for a DSK key"""

    def __init__(self, dsk=None):
        self._dsk = dsk or b""

    def __getstate__(self):
        return "-".join(
            [
                "{:05d}".format((self._dsk[i] << 8) | self._dsk[i + 1])
                for i in range(0, len(self._dsk), 2)
            ]
        )

    def __repr__(self):
        return self.__getstate__()

    def __eq__(self, other):
        return self.__getstate__() == other

    def __setstate__(self, state):
        if isinstance(state, bytes):
            if len(state) == 16 or len(state) == 0:
                self._dsk = state
            else:
                raise ValueError("DSK must be 16 bytes, got {}".format(len(state)))
            return
        fields = state.split("-")
        if len(fields) != 8:
            return
        dsk = bytearray(16)
        for i, value in enumerate(fields):
            word = int(value, base=10)
            dsk[i * 2] = (word >> 8) & 0xFF
            dsk[i * 2 + 1] = word & 0xFF
        self._dsk = bytes(dsk)

    def serialize(self, stream: BitStreamWriter):
        """Serialize DSK"""
        stream.extend(self._dsk)

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize 16 bytes DSK"""
        return cls.deserializeN(stream, 16)

    @staticmethod
    def deserializeN(stream: BitStreamReader, length: int):
        """Deserialize variable length DSK"""
        if length == 0:
            # No key included
            return b""
        if length != 16:
            raise ValueError("DSK must be 16 bytes, got {}".format(length))
        return stream.value(length)


class IPv6(ipaddress.IPv6Address):
    """Type for a IPv6 address"""

    def serialize(self, stream: BitStreamWriter):
        """Serialize the IPv6 address"""
        stream.extend(self.packed)

    @classmethod
    def deserialize(cls, stream: BitStreamReader):
        """Deserialize an IPv6 address"""
        return cls(stream.value(16))


class HomeID(uint32_t):
    """Type for Z-Wave Home ID"""

    def __str__(self):
        return "{:X}".format(self)



# -*- coding: utf-8 -*-

import logging
from typing import Any, Dict

from pyzwave.adapter import Adapter
from pyzwave.commandclass import NetworkManagementInclusion, Zip
from pyzwave.message import Message
from pyzwave.node import Node, NodeEndPoint
from pyzwave.const.ZW_classcmd import (
    GENERIC_TYPE_STATIC_CONTROLLER,
    SPECIFIC_TYPE_GATEWAY,
    COMMAND_CLASS_MULTI_CHANNEL_V2,
)
from pyzwave.persistantstorage import PersistantStorage
from pyzwave.util import Listenable

_LOGGER = logging.getLogger(__name__)


class Application(Listenable):
    """
    Base class for managing the Z-Wave system
    """

    def __init__(self, adapter: Adapter, storage: PersistantStorage):
        super().__init__()
        self.adapter = adapter
        self.adapter.addListener(self)
        self.addListener(storage)
        self._storage = storage
        self._typeInfo = (GENERIC_TYPE_STATIC_CONTROLLER, SPECIFIC_TYPE_GATEWAY)
        self._nodes = {}
        self._cmdClasses = []

    async def loadEndPointNode(self, node: Node, endpoint: int):
        """Load an endpoint for a node"""
        report = await self.adapter.getMultiChannelCapability(node.rootNodeId, endpoint)
        nodeEP = NodeEndPoint(node, endpoint, self.adapter, list(report.commandClass))
        nodeEP.addListener(self._storage)
        nodeEP.genericDeviceClass = report.genericDeviceClass
        nodeEP.specificDeviceClass = report.specificDeviceClass
        self._nodes[nodeEP.nodeId] = nodeEP
        return nodeEP

    async def loadNode(self, nodeId: int) -> list:
        """Load a node"""
        if nodeId == self.adapter.nodeId:
            # Ignore ourself
            return []
        nodes = []
        nodeInfo = await self.adapter.getNodeInfo(nodeId)
        node = Node(nodeId, self.adapter, list(nodeInfo.commandClass))
        nodes.append(node)
        node.addListener(self._storage)
        node.basicDeviceClass = nodeInfo.basicDeviceClass
        # node.flirs = ?
        node.genericDeviceClass = nodeInfo.genericDeviceClass
        # node.isFailed = ?
        node.listening = nodeInfo.listening
        node.specificDeviceClass = nodeInfo.specificDeviceClass
        self._nodes[node.nodeId] = node

        if node.supports(COMMAND_CLASS_MULTI_CHANNEL_V2):
            endpoints = await self.adapter.getMultiChannelEndPoints(nodeId)
            for endpoint in range(1, endpoints + 1):
                nodes.append(await self.loadEndPointNode(node, endpoint))
        return nodes

    async def messageReceived(
        self,
        _sender,
        rootNodeId: int,
        endPoint: int,
        message: Message,
        flags: Zip.HeaderExtension,
    ):
        """Called when a message is received from a node"""
        nodeId = "{}:{}".format(rootNodeId, endPoint)
        node = self._nodes.get(nodeId)
        if node:
            reply = await node.handleMessage(message, flags)
            if reply:
                # Message was handled
                return True
            return False
        if rootNodeId != self.adapter.nodeId:
            # Unknown node. Should not happen
            _LOGGER.warning(
                "Received message from unknown node %s: %s", nodeId, message
            )
            return False
        # Message from the Z-Wave chip
        await self.ask("messageReceived", message)
        # TODO, check retval from above and see if message was handled
        return True

    async def nodeListUpdated(self, _sender):
        """Called when the node list has been updated"""
        nodeList = await self.adapter.getNodeList()
        nodesToRemove = []
        for nodeId, node in self._nodes.items():
            if node.rootNodeId in nodeList:
                continue
            nodesToRemove.append(nodeId)
        for nodeId in nodesToRemove:
            del self._nodes[nodeId]
            await self.ask("nodeRemoved", nodeId)
        # For clients supporting batch removing
        if nodesToRemove:
            await self.ask("nodesRemoved", nodesToRemove)

        # Find new nodes
        newNodes = []
        for nodeId in nodeList:
            if "{}:0".format(nodeId) in self._nodes:
                continue
            newNodes.extend(await self.loadNode(nodeId))

        if newNodes:
            await self.ask("nodesAdded", newNodes)
        for node in newNodes:
            await self.ask("nodeAdded", node)

    @property
    def nodes(self) -> Dict[int, Node]:
        """All nodes in the network"""
        return self._nodes

    async def onMessageReceived(self, _speaker: Any, msg: Message) -> bool:
        """Listener method from the adapter for any unhandled messages"""
        if not isinstance(msg, Zip.ZipPacket):
            # Ignore non ZipPacket
            return False
        command = msg.command
        if isinstance(command, NetworkManagementInclusion.NodeAddStatus):
            await self.ask("addNodeStatus", command)
            return True
        if isinstance(command, NetworkManagementInclusion.NodeRemoveStatus):
            await self.ask("removeNodeStatus", command)
            if (
                command.status
                != NetworkManagementInclusion.NodeRemoveStatus.Status.DONE
            ):
                return True
            if command.nodeID == 0:
                # Fire this case since it may be used to exclude nodes not in our network
                await self.ask("nodeRemoved", 0)
                await self.ask("nodesRemoved", [0])
            return True
        if command is None:
            return False
        for response in await self.ask("messageReceived", command):
            if response is True:
                # It was handled
                return True
        _LOGGER.info("Unhandled message! %s", command.debugString())
        return False

    def setNodeInfo(self, generic, specific, cmdClasses):
        """Set the application NIF (Node Information Frame)"""
        self._typeInfo = (generic, specific)
        self._cmdClasses = cmdClasses

    async def shutdown(self):
        """Shut down the application gracefully"""

    async def startup(self):
        """Start and initialize the application and the adapter"""
        for nodeId in await self.adapter.getNodeList():
            await self.loadNode(nodeId)

        for nodeId in await self.adapter.getFailedNodeList():
            _LOGGER.warning("FAILED node %s", nodeId)

        await self.ask("nodesAdded", list(self._nodes.values()))
        for _, node in self._nodes.items():
            await self.ask("nodeAdded", node)



# -*- coding: utf-8 -*-
import asyncio
from enum import Enum, auto
from contextlib import contextmanager
import logging
from typing import Dict

from pyzwave.adapter import Adapter
from pyzwave.commandclass import (
    CommandClass,
    NetworkManagementInclusion,
    Supervision,
    Zip,
)
from pyzwave.message import Message
from pyzwave.util import Listenable, MessageWaiter

from pyzwave.const.ZW_classcmd import COMMAND_CLASS_ZWAVEPLUS_INFO

_LOGGER = logging.getLogger(__name__)


def supervision(func):
    """Decoratior for handling calls wrapped in supervision messages"""

    async def __wrapper__(self, message: Message, flags):
        if not isinstance(message, Supervision.Get):
            return await func(self, message, flags)
        # Handle supervision calls
        if await func(self, message.command, flags):
            await self.send(
                Supervision.Report(
                    moreStatusUpdates=False,
                    wakeUpRequest=False,
                    sessionID=message.sessionID,
                    status=0xFF,
                    duration=0,
                )
            )
            return True
        await self.send(
            Supervision.Report(
                moreStatusUpdates=False,
                wakeUpRequest=False,
                sessionID=message.sessionID,
                status=0x00,
                duration=0,
            )
        )
        return False

    return __wrapper__


class StorageStatus(Enum):
    """Storage status flag"""

    CLEAN = auto()
    LOCKED_CLEAN = auto()
    LOCKED_DIRTY = auto()


class Node(Listenable, MessageWaiter):
    """
    Base class for a Z-Wave node
    """

    def __init__(self, nodeId: int, adapter: Adapter, cmdClasses: list):
        super().__init__()
        self._adapter = adapter
        self._basicDeviceClass = 0
        self._controlled = {}
        # Flag is this node need to be saved to persistant storage
        self._storageStatus: StorageStatus = StorageStatus.CLEAN
        self._endpoint = 0
        self._flirs = False
        self._genericDeviceClass = 0
        self._isFailed = False
        self._listening = False
        self._nodeId = nodeId
        self._specificDeviceClass = 0
        self._supported = {}
        i = 0
        supported = True
        securityS0 = False
        while i < len(cmdClasses):
            cmdClass = cmdClasses[i]
            if cmdClass == 0xF1 and cmdClasses[i + 1] == 0x00:
                # Security Scheme 0 Mark
                securityS0 = True
                supported = True
                i += 2
                continue
            if cmdClass == 0xEF:
                # Support/Control mark
                supported = False
                i += 1
                continue
            clsObject = CommandClass.load(cmdClass, securityS0, self)
            if supported:
                self._supported[cmdClass] = clsObject
                clsObject.addListener(self)
            else:
                self._controlled[cmdClass] = clsObject
            # _LOGGER.info(
            #     "Load command class for %X %s", cmdClass, clsObject,
            # )
            i += 1

    @property
    def adapter(self) -> Adapter:
        """The adapter"""
        return self._adapter

    @property
    def basicDeviceClass(self) -> int:
        """Return this nodes basic device class"""
        return self._basicDeviceClass

    @basicDeviceClass.setter
    def basicDeviceClass(self, basicDeviceClass: int):
        self._basicDeviceClass = basicDeviceClass

    def commandClassUpdated(self, _commandClass: CommandClass):
        """Called by the command classes if their data is updated"""
        if self._storageStatus == StorageStatus.CLEAN:
            # Signal update needed
            self.speak("nodeUpdated")
        elif self._storageStatus == StorageStatus.LOCKED_CLEAN:
            # Storage is locked, move to dirty
            self._storageStatus = StorageStatus.LOCKED_DIRTY

    @property
    def endpoint(self) -> int:
        """Returns the endpoint if this is a subnode. 0 if it is the root node"""
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint: int):
        self._endpoint = endpoint

    @property
    def flirs(self) -> bool:
        """Returns if this node is a FLiRS node or not"""
        return self._flirs

    @flirs.setter
    def flirs(self, flirs: bool):
        self._flirs = flirs

    @property
    def genericDeviceClass(self) -> int:
        """Returns this nodes generic device class"""
        return self._genericDeviceClass

    @genericDeviceClass.setter
    def genericDeviceClass(self, genericDeviceClass: int):
        self._genericDeviceClass = genericDeviceClass

    @supervision
    async def handleMessage(self, message: Message, flags: Zip.HeaderExtension) -> bool:
        """Handle and incomming message. Route it to the correct handler"""
        handled = False
        if self.messageReceived(message) is True:
            # Message has already been handled
            handled = True
        cmdClass: CommandClass = self.supported.get(message.cmdClass())
        if cmdClass:
            retval = await cmdClass.handleMessage(message, flags)
            if retval:
                return retval
        if handled:
            return True
        for retval in await self.ask("onMessage", message):
            if retval:
                return retval
        # Message was not handled
        _LOGGER.warning("Unhandled message %s from node %s", message, self.nodeId)
        _LOGGER.debug(message.debugString())
        return False

    async def interview(self):
        """
        (Re)interview this node. It is recommended to apply the :meth:`storageLock()`
        before calling this function.
        """
        for _, cmdClass in self._supported.items():
            try:
                await cmdClass.interview()
            except asyncio.TimeoutError:
                _LOGGER.warning("Timeout interviewing %s", cmdClass.name)

    @property
    def isFailed(self) -> bool:
        """Returns is this node is considered failing or not"""
        return self._isFailed

    @isFailed.setter
    def isFailed(self, isFailed: bool):
        self._isFailed = isFailed

    @property
    def isZWavePlus(self) -> bool:
        """Returns True if this is a Z-Wave Plus node"""
        return self.supports(COMMAND_CLASS_ZWAVEPLUS_INFO)

    @property
    def listening(self) -> bool:
        """Returns True if this node is listening. False if it is a sleeping node"""
        return self._listening

    @listening.setter
    def listening(self, listening: bool):
        self._listening = listening

    @property
    def nodeId(self) -> str:
        """The node id in the format nodeid:channel"""
        return "{}:{}".format(self._nodeId, self._endpoint)

    async def remove(self) -> bool:
        """
        Remove this node from the network. This only works if the node is
        marked as failed in the Z-Wave chip
        """
        status = await self.adapter.removeFailedNode(self.rootNodeId)
        if status == NetworkManagementInclusion.FailedNodeRemoveStatus.Status.DONE:
            return True
        _LOGGER.warning("Could not remove failed node, reason %s", status)
        return False

    @property
    def rootNodeId(self) -> int:
        """Return the root node id"""
        return self._nodeId

    async def send(self, cmd: Message, timeout: int = 3) -> bool:
        """Send a message to this node"""
        return await self._adapter.sendToNode(
            self._nodeId, cmd, sourceEP=0, destEP=self._endpoint, timeout=timeout
        )

    async def sendAndReceive(
        self, cmd: Message, waitFor: Message, timeout: int = 3, **kwargs
    ) -> Message:
        """Send a message and wait for the response"""
        self.addWaitingSession(waitFor)
        retval = await self.send(cmd, **kwargs)
        if not retval:
            # Node did not wakeup in time!
            raise asyncio.TimeoutError()
        return await self.waitForMessage(waitFor, timeout=timeout)

    @contextmanager
    def storageLock(self):
        """
        Suppresses (temporarily) the signals to store settings persistant.

        Some memories do not like to be written to often, such as flash memories
        found in embedded boards. If the application knows the node will be updated
        a lock can be added so it will only be written to disc once all operations has
        been finished.
        To use this lock use the with-statement:

        .. code-block:: python

            with node.storageLock():
                # Do operations with the node here.
                # Nothing will be stored to disk.
                node.interview()
            # The storage will happen here, only once
        """
        try:
            self._storageStatus = StorageStatus.LOCKED_CLEAN
            yield
        finally:
            if self._storageStatus == StorageStatus.LOCKED_DIRTY:
                self.speak("nodeUpdated")
            self._storageStatus = StorageStatus.CLEAN

    @property
    def specificDeviceClass(self) -> int:
        """Returns this nodes specific device class"""
        return self._specificDeviceClass

    @specificDeviceClass.setter
    def specificDeviceClass(self, specificDeviceClass: int):
        self._specificDeviceClass = specificDeviceClass

    @property
    def supported(self) -> Dict[int, CommandClass]:
        """Return a dict of command classes this node supports"""
        return self._supported

    def supports(self, commandClass: int) -> bool:
        """Returns if this node supports a specific command class or not"""
        return commandClass in self._supported


class NodeEndPoint(Node):
    """Base class for a sub node for nodes supporting command class multi channel"""

    def __init__(self, parent: Node, endpoint: int, adapter: Adapter, cmdClasses: list):
        super().__init__(parent.rootNodeId, adapter, cmdClasses)
        self._parent = parent
        self.endpoint = endpoint

    @property
    def basicDeviceClass(self) -> int:
        """Return this nodes basic device class"""
        return self._parent.basicDeviceClass

    @property
    def flirs(self) -> bool:
        """Returns if this node is a FLiRS node or not"""
        return self._parent.flirs

    @property
    def isFailed(self) -> bool:
        """Returns is this node is considered failing or not"""
        return self._parent.isFailed

    @property
    def listening(self) -> bool:
        """Returns True if this node is listening. False if it is a sleeping node"""
        return self._parent.listening

    @property
    def parent(self) -> Node:
        """Returns the parent node"""
        return self._parent



# -*- coding: utf-8 -*-

import asyncio
import ipaddress
import logging

from pyzwave.message import Message
from pyzwave.commandclass import (
    NetworkManagementInclusion,
    NetworkManagementProxy,
    Zip,
    ZipGateway,
    ZipND,
)
from pyzwave.adapter import TxOptions
from pyzwave.connection import Connection
from pyzwave.zipconnection import ZIPConnection
from pyzwave.types import dsk_t

_LOGGER = logging.getLogger(__name__)


class ZIPGateway(ZIPConnection):
    """Class for communicating with a zipgateway"""

    def __init__(self, address, psk):
        super().__init__(address, psk)
        self._unsolicitedConnection = Connection()
        self._unsolicitedConnection.onMessage(self.onUnsolicitedMessage)
        self._connections = {}
        self._nodes = {}
        self._nmSeq = 0

    async def addNode(self, txOptions: TxOptions) -> bool:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.NodeAdd(
            seqNo=self._nmSeq,
            mode=NetworkManagementInclusion.NodeAdd.Mode.ANY_S2,
            txOptions=txOptions,
        )
        try:
            # We do not get any immediate response
            await self.send(cmd)
        except asyncio.TimeoutError:
            return False
        return True

    async def addNodeDSKSet(
        self, accept: bool, inputDSKLength: int, dsk: dsk_t
    ) -> bool:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.NodeAddDSKSet(
            seqNo=self._nmSeq, accept=accept, inputDSKLength=inputDSKLength, dsk=dsk,
        )
        try:
            await self.send(cmd)
        except asyncio.TimeoutError:
            return False
        return True

    async def addNodeKeysSet(
        self, grantCSA: bool, accept: bool, grantedKeys: NetworkManagementInclusion.Keys
    ) -> bool:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.NodeAddKeysSet(
            seqNo=self._nmSeq,
            grantCSA=grantCSA,
            accept=accept,
            grantedKeys=grantedKeys,
        )
        try:
            await self.send(cmd)
        except asyncio.TimeoutError:
            return False
        return True

    async def addNodeStop(self) -> bool:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.NodeAdd(
            seqNo=self._nmSeq,
            mode=NetworkManagementInclusion.NodeAdd.Mode.STOP,
            txOptions=TxOptions.NULL,
        )
        try:
            await self.send(cmd)
        except asyncio.TimeoutError:
            return False
        return True

    async def connect(self):
        await super().connect()
        await self.setGatewayMode(1)
        await self.setupUnsolicitedConnection()

    async def connectToNode(self, nodeId) -> ZIPConnection:
        """Returns a connection to the node"""
        if nodeId in self._connections:
            return self._connections[nodeId]

        ipv6 = await self.ipOfNode(nodeId)

        connection = ZIPConnection(ipv6.compressed, self.psk)
        connection.addListener(self)
        await connection.connect()
        self._connections[nodeId] = connection
        return connection

    async def getFailedNodeList(self) -> list:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementProxy.FailedNodeListGet(seqNo=self._nmSeq,)
        try:
            report = await self.sendAndReceive(
                cmd, NetworkManagementProxy.FailedNodeListReport
            )
        except asyncio.TimeoutError:
            return set()
        return report.failedNodeList

    async def getMultiChannelEndPoints(self, nodeId: int) -> int:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementProxy.MultiChannelEndPointGet(
            seqNo=self._nmSeq, nodeID=nodeId
        )
        try:
            report = await self.sendAndReceive(
                cmd, NetworkManagementProxy.MultiChannelEndPointReport
            )
        except asyncio.TimeoutError:
            # No response
            return 0
        return report.individualEndPoints + report.aggregatedEndPoints

    async def getMultiChannelCapability(
        self, nodeId: int, endpoint: int
    ) -> NetworkManagementProxy.MultiChannelCapabilityReport:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementProxy.MultiChannelCapabilityGet(
            seqNo=self._nmSeq, nodeID=nodeId, endPoint=endpoint
        )
        try:
            report = await self.sendAndReceive(
                cmd, NetworkManagementProxy.MultiChannelCapabilityReport
            )
        except asyncio.TimeoutError:
            # No response
            return NetworkManagementProxy.MultiChannelCapabilityReport()
        return report

    async def getNodeList(self) -> set:
        if self._nodes:
            # Return cached list
            return set(self._nodes.keys())
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementProxy.NodeListGet(seqNo=self._nmSeq)
        try:
            report = await self.sendAndReceive(
                cmd, NetworkManagementProxy.NodeListReport
            )
        except asyncio.TimeoutError:
            # No response
            return set()
        self._nodeId = report.nodeListControllerId
        self._nodes = {x: {} for x in report.nodes}
        return report.nodes

    async def getNodeInfo(
        self, nodeId: int
    ) -> NetworkManagementProxy.NodeInfoCachedReport:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementProxy.NodeInfoCachedGet(
            seqNo=self._nmSeq, maxAge=15, nodeID=nodeId
        )
        try:
            report = await self.sendAndReceive(
                cmd, NetworkManagementProxy.NodeInfoCachedReport
            )
        except asyncio.TimeoutError:
            return NetworkManagementProxy.NodeInfoCachedReport()
        return report

    async def __handleNodeListReport__(
        self, report: NetworkManagementProxy.NodeListReport
    ):
        """Update nodes from a nodeListReport"""
        # Find removed nodes
        for nodeId in list(self._nodes.keys()):
            if nodeId in report.nodes:
                continue
            del self._nodes[nodeId]

        # Find new nodes
        for nodeId in report.nodes:
            if nodeId in self._nodes:
                continue
            # Retrieve ip
            self._nodes[nodeId] = {"ip": await self.ipOfNode(nodeId)}
        self.speak("nodeListUpdated")

    async def ipOfNode(self, nodeId) -> ipaddress.IPv6Address:
        """Returns the IPv6 address of the node"""
        msg = ZipND.ZipInvNodeSolicitation(local=False, nodeId=nodeId)
        self._conn.send(msg.compose())
        response = await self.waitForMessage(ZipND.ZipNodeAdvertisement, timeout=3)
        return response.ipv6

    async def removeFailedNode(
        self, nodeId: int
    ) -> NetworkManagementInclusion.FailedNodeRemoveStatus.Status:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.FailedNodeRemove(
            seqNo=self._nmSeq, nodeID=nodeId
        )
        try:
            # The timeout might need to be increased. The Z-Wave may need to query the
            # network for a response first and this may take time on a large network
            report = await self.sendAndReceive(
                cmd, NetworkManagementInclusion.FailedNodeRemoveStatus, timeout=10
            )
        except asyncio.TimeoutError:
            return NetworkManagementInclusion.FailedNodeRemoveStatus.Status.REMOVE_FAIL
        return report.status

    async def removeNode(self) -> bool:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.NodeRemove(
            seqNo=self._nmSeq, mode=NetworkManagementInclusion.NodeRemove.Mode.ANY
        )
        try:
            # We do not get any immediate response
            await self.send(cmd)
        except asyncio.TimeoutError:
            return False
        return True

    async def removeNodeStop(self) -> bool:
        self._nmSeq = self._nmSeq + 1
        cmd = NetworkManagementInclusion.NodeRemove(
            seqNo=self._nmSeq, mode=NetworkManagementInclusion.NodeRemove.Mode.STOP
        )
        try:
            # We do not get any immediate response
            await self.send(cmd)
        except asyncio.TimeoutError:
            return False
        return True

    def onMessageReceived(self, connection: ZIPConnection, message: Message):
        """Called when a message is received from any node connection. Not unsolicited."""
        sourceEP = 0
        if isinstance(message, Zip.ZipPacket):
            sourceEP = message.sourceEP
            message = message.command
        for nodeId, nodeConnection in self._connections.items():
            if connection != nodeConnection:
                continue
            flags = 0  # Set flags such as encapsulation type
            self.speak("messageReceived", nodeId, sourceEP, message, flags)
            return
        _LOGGER.warning("Got message from unknown sender %s: %s", connection, message)

    def onUnsolicitedMessage(self, pkt, address):
        """
        Called when an unsolicited message is received.
        We do not know the node id the message is from. Only the ip address.
        """
        zipPkt = Message.decode(pkt)
        sourceIp = ipaddress.IPv6Address(address[0])
        sourceEP = zipPkt.sourceEP

        if isinstance(zipPkt.command, NetworkManagementProxy.NodeListReport):
            return asyncio.ensure_future(self.__handleNodeListReport__(zipPkt.command))

        # Find the node this was from
        for nodeId, node in self._nodes.items():
            if node.get("ip") != sourceIp:
                continue
            if zipPkt.ackRequest:
                # This message needs an ack response.
                ackReponse = zipPkt.response(success=True)
                self._unsolicitedConnection.sendTo(ackReponse.compose(), address)

            self.speak(
                "messageReceived",
                nodeId,
                sourceEP,
                zipPkt.command,
                zipPkt.headerExtension,
            )
            return True
        _LOGGER.warning(
            "Got message from unknown sender %s: %s", sourceIp, zipPkt.command
        )
        return False

    async def sendToNode(self, nodeId: int, cmd: Message, **kwargs) -> bool:
        conn = await self.connectToNode(nodeId)
        return await conn.send(cmd, **kwargs)

    async def setGatewayMode(self, mode: int, timeout: int = 3) -> bool:
        """Set gateway to standalone or portal mode"""
        try:
            report = await self.sendAndReceive(
                ZipGateway.GatewayModeGet(),
                ZipGateway.GatewayModeReport,
                timeout=timeout,
            )
            if report.mode != mode:
                await self.send(ZipGateway.GatewayModeSet(mode=mode))
        except asyncio.TimeoutError:
            _LOGGER.error("Got timeout waiting for gateway mode report")
            return False
        return True

    async def setNodeInfo(self, generic, specific, cmdClasses):
        raise NotImplementedError()

    async def setupUnsolicitedConnection(self):
        """
        Setup for listening for unsolicited connections. This function must not be called
        explicitly. It is called by the connect() method automatically
        """
        await self._unsolicitedConnection.listen(self.psk, 4123)
        await self.send(
            ZipGateway.UnsolicitedDestinationSet(
                unsolicitedIPv6Destination=ipaddress.IPv6Address("::ffff:c0a8:31"),
                port=4123,
            )
        )
        # Retrieve node list, and get ip-addresses
        for nodeId in await self.getNodeList():
            if not self._nodes[nodeId].get("ip"):
                self._nodes[nodeId] = {"ip": await self.ipOfNode(nodeId)}



# -*- coding: utf-8 -*-

import asyncio
import logging

_LOGGER = logging.getLogger(__name__)


class ZipClientProtocol:
    """Internal ZIP Client protocol implementation"""

    def __init__(self, onConLost, onMessage):
        self.onConLost = onConLost
        self.onMessage = onMessage
        self.transport = None

    def connection_made(self, transport):  # pylint: disable=invalid-name
        """Called when a new connection is made"""
        self.transport = transport

    def datagram_received(self, data, addr):  # pylint: disable=invalid-name
        """Called when a new udp packet has arrived"""
        self.onMessage(data, addr)

    def error_received(self, exc):  # pylint: disable=invalid-name,no-self-use
        """Called when error happens"""
        _LOGGER.error("Error received: %s", exc)

    def connection_lost(self, exc):  # pylint: disable=invalid-name
        """Called when connection is lost"""
        _LOGGER.warning("Connection closed %s", exc)
        self.onConLost.set_result(True)


class Connection:
    """Connection object to create a non encrypted connection using PSK"""

    def __init__(self):
        super().__init__()
        self._address = None
        self._psk = None
        self._running = False
        self._sock = None
        self._onMessage = None
        self._server = False

    async def connect(self, address, psk):
        """Connect to remote using psk"""
        self._address = address
        self._psk = psk

        loop = asyncio.get_event_loop()
        onConLost = loop.create_future()
        self._sock, _protocol = await loop.create_datagram_endpoint(
            lambda: ZipClientProtocol(onConLost, self._msgReceived),
            remote_addr=(self._address, 4123),
        )
        asyncio.ensure_future(self.run(onConLost))

    async def listen(self, psk, port):
        """Start server socket"""
        self._server = True
        self._psk = psk
        loop = asyncio.get_event_loop()
        onConLost = loop.create_future()
        self._sock, _protocol = await loop.create_datagram_endpoint(
            lambda: ZipClientProtocol(onConLost, self._msgReceived),
            local_addr=("::", port),
        )

    async def run(self, onConLost):  # pylint: disable=missing-function-docstring
        try:
            await onConLost
        finally:
            self._sock.close()

    def send(self, msg) -> bool:
        """Send bytes to socket"""
        if not self._sock:
            _LOGGER.error("Could not send, not yet connected!")
            return False
        self._sock.sendto(msg)
        return True

    def sendTo(self, msg, address) -> bool:
        """Send bytes to address"""
        self._sock.sendto(msg, address)

    def onMessage(self, cbfn):
        """Set the callback function to use when data has arrived"""
        self._onMessage = cbfn

    def stop(self):
        """Stop the thread"""
        self._running = False
        self._sock.close()

    def _msgReceived(self, msg, sender):
        if not self._onMessage:
            return
        if self._server:
            self._onMessage(msg, sender)
        else:
            self._onMessage(msg)



# -*- coding: utf-8 -*-

import logging
import pathlib
import yaml

from pyzwave.commandclass import CommandClass
from pyzwave.node import Node

_LOGGER = logging.getLogger(__name__)


class PersistantStorage:
    """Base class for implementing persistant storage for nodes"""

    def nodeAdded(self, application, node: Node):
        """Called when a new node has been added and/or loaded"""

    def nodeUpdated(self, node: Node):
        """
        Called then the settings for a node or one of it's command classes
        has been updated
        """


class YamlStorage(PersistantStorage):
    """Store persistant settings as yaml files"""

    def __init__(self, path):
        self._path = path

    def nodeAdded(self, application, node: Node):
        path: pathlib.Path = self.pathForNode(node.nodeId)
        if not path.exists():
            return
        with open(path, "r") as fd:
            # We use base loader here to not let yaml initiate command classes itself
            settings = yaml.load(fd, Loader=yaml.BaseLoader)
        supported = settings.get("supported", {})
        for cmdClassId, cmdClass in node.supported.items():
            if str(cmdClassId) not in supported:
                continue
            cmdClass.__setstate__(supported[str(cmdClassId)])

    def nodeUpdated(self, node: Node):
        settings = {"supported": {}}
        for cmdClassId, cmdClass in node.supported.items():
            settings["supported"][cmdClassId] = self.cmdClassRepresenter(cmdClass)
        with open(self.pathForNode(node.nodeId), "w") as fd:
            yaml.dump(settings, fd)

    @property
    def path(self) -> str:
        """The path to store the yaml files"""
        return self._path

    @path.setter
    def path(self, path: str):
        self._path = path

    def pathForNode(self, nodeId: int) -> pathlib.Path:
        """Returns the path for settings for a node"""
        return pathlib.Path(self._path) / "nodes" / "{}.yml".format(nodeId)

    @staticmethod
    def cmdClassRepresenter(cmdClass: CommandClass):
        """Wrapper method for converting to YAML format"""

        class YAMLCommandClass(
            yaml.YAMLObject
        ):  # pylint: disable=missing-class-docstring
            yaml_tag = cmdClass.NAME

            def __getstate__(self):
                return cmdClass.__getstate__()

        return YAMLCommandClass()



# -*- coding: utf-8 -*-

import asyncio
import logging

from pyzwave.message import Message
from pyzwave.commandclass import (
    NetworkManagementInclusion,
    NetworkManagementProxy,
    Zip,
    ZipND,
)
from pyzwave.connection import Connection
from .adapter import Adapter, TxOptions
from .types import dsk_t

_LOGGER = logging.getLogger(__name__)


class ZIPConnection(Adapter):
    """Class for connecting to a zipgateway or zipnode"""

    def __init__(self, address, psk):
        super().__init__()
        self._seq = 0
        self._address = address
        self._keepAlive = None
        self._psk = psk
        self._conn = Connection()
        self._conn.onMessage(self.onPacket)

    async def addNode(self, txOptions: TxOptions) -> bool:
        raise NotImplementedError()

    async def addNodeDSKSet(
        self, accept: bool, inputDSKLength: int, dsk: dsk_t
    ) -> bool:
        raise NotImplementedError()

    async def addNodeKeysSet(
        self, grantCSA: bool, accept: bool, grantedKeys: NetworkManagementInclusion.Keys
    ) -> bool:
        raise NotImplementedError()

    async def addNodeStop(self) -> bool:
        raise NotImplementedError()

    async def connect(self):
        await self._conn.connect(self._address, self._psk)
        self.resetKeepAlive()

    async def getFailedNodeList(self) -> list:
        raise NotImplementedError()

    async def getMultiChannelCapability(
        self, nodeId: int, endpoint: int
    ) -> NetworkManagementProxy.MultiChannelCapabilityReport:
        raise NotImplementedError()

    async def getMultiChannelEndPoints(self, nodeId: int) -> int:
        raise NotImplementedError()

    async def getNodeList(self) -> set:
        raise NotImplementedError()

    async def getNodeInfo(
        self, nodeId: int
    ) -> NetworkManagementProxy.NodeInfoCachedReport:
        raise NotImplementedError()

    def keepAlive(self):
        """Send a keepalive message"""
        msg = Zip.ZipKeepAlive(ackRequest=True, ackResponse=False,)
        self._conn.send(msg.compose())
        self.resetKeepAlive()

    def onPacket(self, pkt):
        """Called when a packed has recevied from the connection"""
        try:
            zipPkt = Message.decode(pkt)
        except Exception:
            _LOGGER.error("Could not decode message. Raw message:")
            _LOGGER.error("%s", pkt)
            return False
        if isinstance(zipPkt, Zip.ZipPacket):
            if zipPkt.ackResponse:
                self.ackReceived(zipPkt)
                return True
            if zipPkt.nackResponse:
                if zipPkt.nackWaiting:
                    # Waiting: the preceding Z/IP Packet encapsulated Z-Wave Command is not yet
                    # delivered to the destination and delivery will be attempted later on
                    # Threat these as normal acks
                    self.ackReceived(zipPkt)
                    return True
                _LOGGER.error("Nack response not implemented %s", zipPkt.debugString())
                # self.nackReceived(zipPkt.seqNo)
                return False
            if zipPkt.ackRequest:
                _LOGGER.error("This message needs an ack response. Not implemented")
                return False
            if zipPkt.zwCmdIncluded:
                self.commandReceived(zipPkt)
        elif isinstance(zipPkt, Zip.ZipKeepAlive):
            if zipPkt.ackResponse:
                # Ignore a response
                return True
            _LOGGER.error("This message needs an ack response. Not implemented")
            return False
        elif isinstance(zipPkt, ZipND.ZipNodeAdvertisement):
            self.commandReceived(zipPkt)
        else:
            _LOGGER.warning("Received unknown Z/IP packet from zipgateway: %s", zipPkt)
            return False
        return True

    @property
    def psk(self) -> bytes:
        """The psk used for the connection"""
        return self._psk

    async def removeFailedNode(
        self, nodeId: int
    ) -> NetworkManagementInclusion.FailedNodeRemoveStatus.Status:
        raise NotImplementedError()

    async def removeNode(self) -> bool:
        raise NotImplementedError()

    async def removeNodeStop(self) -> bool:
        raise NotImplementedError()

    def resetKeepAlive(self):
        """Reset the keepalive timeout"""
        if self._keepAlive:
            # Timer running, reset it
            self._keepAlive.cancel()
        self._keepAlive = asyncio.get_event_loop().call_later(25, self.keepAlive)

    async def send(self, cmd, sourceEP=0, destEP=0, timeout=3) -> bool:
        self._seq = (self._seq + 1) & 0xFF
        msg = Zip.ZipPacket(
            ackRequest=True,
            ackResponse=False,
            nackResponse=False,
            nackWaiting=False,
            nackQueueFull=False,
            nackOptionError=False,
            headerExtIncluded=False,
            zwCmdIncluded=True,
            moreInformation=False,
            secureOrigin=True,
            seqNo=self._seq,
            sourceEP=sourceEP,
            destEP=destEP,
            command=cmd,
        )
        self._conn.send(msg.compose())
        self.resetKeepAlive()
        try:
            await self.waitForAck(msg.seqNo, timeout=timeout)
        except asyncio.TimeoutError:
            return False
        return True

    async def setNodeInfo(self, generic, specific, cmdClasses):
        raise NotImplementedError()



import logging

from pyzwave.types import BitStreamReader, BitStreamWriter
from pyzwave.util import AttributesMixin

_LOGGER = logging.getLogger(__name__)


class Message(AttributesMixin):
    """Base class for all Z-Wave messages. This class should not be initiated manually"""

    NAME = None

    def cmdClass(self) -> int:
        """Return the command class id for this message"""
        return (self.hid() >> 8) & 0xFF

    def compose(self) -> bytes:
        """Convert the message to a bytearray ready to be sent over the wire"""
        cmdClass, cmd = ZWaveMessage.reverseMapping.get(self.__class__, (None, None))
        if cmdClass is None or cmd is None:
            # Pure Message object, encode to empty bytes
            return b""
        stream = BitStreamWriter()
        stream.addBytes(cmdClass, 1, False)
        stream.addBytes(cmd, 1, False)
        for name, attrType in self.attributes:
            if name not in self._attributes:
                default = getattr(attrType, "default", None)
                if default is not None:
                    attrType(default).serialize(stream)
                    continue
                raise AttributeError(
                    "Value for attribute '{}' in {} has not been set".format(name, self)
                )
            serializer = getattr(self, "compose_{}".format(name), None)
            if serializer:
                serializer(stream)
                continue
            if not hasattr(self._attributes[name], "serialize"):
                raise ValueError("Cannot encode", name, self._attributes[name])
            self._attributes[name].serialize(stream)
        return bytes(stream)

    def serialize(self, stream: BitStreamWriter):
        """Write the message as binary into the bitstream. See compose()"""
        stream.extend(self.compose())

    def __eq__(self, other):
        if not isinstance(other, Message):
            return False
        return self.compose() == other.compose()

    def __getattr__(self, name):
        # Default implentation in AttributesMixin returns (and sets) a default value.
        # We do not want it here. We only return default if it is explicit set.
        if name not in self._attributes:
            for attrName, attrType in getattr(self, "attributes"):
                if attrName != name:
                    continue
                return getattr(attrType, "default", None)
        return self._attributes.get(name)

    def __repr__(self):
        hid = self.hid()
        cmdClass = (hid >> 8) & 0xFF
        cmd = hid & 0xFF
        cmdClassName = cmdClasses.get(cmdClass, "0x{:02X}".format(cmdClass))
        name = self.NAME or "0x{:02X}".format(cmd)
        return "<Z-Wave {}.{}>".format(cmdClassName, name)

    @classmethod
    def decode(cls, pkt: bytearray):
        """Decode a raw bytearray into a Message object"""
        return cls.deserialize(BitStreamReader(pkt))

    @staticmethod
    def deserialize(stream: BitStreamReader):
        """Deserialize a bitstream into a Message object"""
        if stream.bytesLeft() < 2:
            return UnknownMessage(0x0000)
        cmdClass = stream.byte()
        cmd = stream.byte()
        hid = cmdClass << 8 | (cmd & 0xFF)
        MsgCls = ZWaveMessage.get(hid, None)  # pylint: disable=invalid-name
        if MsgCls:
            msg = MsgCls()
            msg.parseAttributes(stream)
        else:
            msg = UnknownMessage(hid)
        return msg

    @classmethod
    def hid(cls):
        """
        Return the command class id and command id as a single word for easier lookup
        """
        cmdClass, cmd = ZWaveMessage.reverseMapping.get(cls, (0, 0))
        return (cmdClass << 8) | (cmd & 0xFF)


class UnknownMessage(Message):
    """
    Wrapper class for wrapping unknown messages. Using this class we still know which
    command class and command this message is (but we don't know how to decode it)
    """

    def __init__(self, hid):
        super().__init__()
        self._hid = hid

    def hid(self):
        return self._hid


# pylint: disable=wrong-import-position
from pyzwave.commandclass import ZWaveMessage, cmdClasses



# -*- coding: utf-8 -*-
import asyncio
from ipaddress import IPv6Address
import logging

import crcmod.predefined

from pyzwave.adapter import Adapter
from pyzwave.commandclass import Mailbox, Zip
from pyzwave.message import Message

_LOGGER = logging.getLogger(__name__)


class QueueItem:
    """
    Class for holding one queue entry. It is also responsible for sending
    it's heartbeats
    """

    def __init__(self, nodeId: int, handle: int, data: bytes, adapter):
        self._adapter = adapter
        self._nodeId = nodeId
        self._handle = handle
        self._data = data
        self._task = None

    @property
    def checksum(self) -> int:
        """Return a checksum of the data"""
        # From the docs
        # To avoid duplicate entries, the Mailbox Service MUST maintain a list of CRC16
        # checksums for each mailbox entry. All mailbox entries MUST be unique, if a
        # matching CRC16 exists for an incoming package, the incoming package MUST be
        # discarded.
        crc16 = crcmod.predefined.mkCrcFun("crc-aug-ccitt")
        return crc16(self._data)

    @property
    def data(self) -> bytes:
        """The qctual queue data"""
        return self._data

    def start(self):
        """Start the task for sending heartbeats to the mailbox proxy"""
        self._task = asyncio.ensure_future(self._runner())

    def stop(self):
        """Stop sending heartbeats"""
        if self._task:
            self._task.cancel()

    async def _runner(self):
        i = 0
        while True:
            i += 1
            await asyncio.sleep(60)
            operation = Mailbox.Queue.Operation.WAITING
            if i % 10 == 0:
                operation = Mailbox.Queue.Operation.PING
            await self._adapter.sendToNode(
                self._nodeId,
                Mailbox.Queue(
                    last=False,
                    operation=operation,
                    queueHandle=self._handle,
                    mailboxEntry=self._data,
                ),
            )


class MailboxService:
    """Mailbox for storing messages for sleeping nodes"""

    def __init__(self, adapter):
        self._adapter: Adapter = adapter
        self._adapter.addListener(self)
        self._lastQueueId = None
        self._queue = {}

    async def initialize(self, ipaddress: IPv6Address, port: int) -> bool:
        """Initialize the mailbox"""
        msg = Mailbox.ConfigurationSet(
            mode=Mailbox.Mode.ENABLE_MAILBOX_PROXY_FORWARDING,
            forwardingDestinationIPv6Address=ipaddress,
            udpPortNumber=port,
        )
        return await self._adapter.send(msg)

    async def __popQueue__(self, nodeId: int, queueHandle: int):
        _LOGGER.debug("Pop queue (%d)", len(self._queue.get(queueHandle, [])))
        if not self._queue.get(queueHandle, None):
            return await self._adapter.sendToNode(
                nodeId,
                Mailbox.Queue(
                    last=True,
                    operation=Mailbox.Queue.Operation.POP,
                    queueHandle=queueHandle,
                ),
            )
        queueItem = self._queue[queueHandle].pop()
        ret = await self._adapter.sendToNode(
            nodeId,
            Mailbox.Queue(
                last=len(self._queue[queueHandle]) == 0,
                operation=Mailbox.Queue.Operation.POP,
                queueHandle=queueHandle,
                mailboxEntry=queueItem.data,
            ),
        )
        _LOGGER.info("Sent message from queue: %s", ret)
        if ret:
            queueItem.stop()
        return True

    async def __pushQueue__(self, nodeId: int, queueHandle: int, message: bytes):
        queueItem = QueueItem(nodeId, queueHandle, message, self._adapter)
        checksum = queueItem.checksum
        for item in self._queue.setdefault(queueHandle, []):
            if item.checksum == checksum:
                # Message has already been queueud
                return False
        ret = await self._adapter.sendToNode(
            nodeId,
            Mailbox.Queue(
                last=False,
                operation=Mailbox.Queue.Operation.WAITING,
                queueHandle=queueHandle,
                mailboxEntry=message,
            ),
        )
        if ret:
            queueItem.start()
            self._queue[queueHandle].append(queueItem)
        return ret

    async def messageReceived(
        self,
        _sender,
        rootNodeId: int,
        _endPoint: int,
        message: Message,
        _flags: Zip.HeaderExtension,
    ):
        """Handle incoming mailbox messages"""
        if isinstance(message, Mailbox.WakeupNotification):
            self._lastQueueId = message.queueHandle
            return await self.__popQueue__(rootNodeId, message.queueHandle)

        if (
            isinstance(message, Mailbox.Queue)
            and message.operation == Mailbox.Queue.Operation.ACK
        ):
            queueHandle = message.queueHandle or self._lastQueueId
            return await self.__popQueue__(rootNodeId, queueHandle)

        if (
            isinstance(message, Mailbox.Queue)
            and message.operation == Mailbox.Queue.Operation.PUSH
        ):
            return await self.__pushQueue__(
                rootNodeId, message.queueHandle, message.mailboxEntry
            )



# -*- coding: utf-8 -*-

import abc
import asyncio
import enum
import logging

from pyzwave.commandclass import NetworkManagementInclusion, NetworkManagementProxy, Zip
from .util import Listenable, MessageWaiter
from .types import dsk_t
from .message import Message

_LOGGER = logging.getLogger(__name__)


class TxOptions(enum.IntFlag):
    """TX Options used for adding nodes to network"""

    NULL = 0x00
    TRANSMIT_OPTION_LOW_POWER = 0x02
    TRANSMIT_OPTION_EXPLORE = 0x20


class Ack:
    """Class for holding session informaion"""

    class Status(enum.Enum):
        """Ack status"""

        PENDING = enum.auto()
        QUEUED = enum.auto()
        RECEIVED = enum.auto()

    def __init__(self):
        self._event = asyncio.Event()
        self._status = Ack.Status.PENDING
        self._expectedDelay = 0

    def received(self):
        """Call this function when this ack has been received"""
        self._status = Ack.Status.RECEIVED
        self._event.set()

    def queued(self, expectedDelay: int):
        """Call this function when the message cannot be delivered right now"""
        if expectedDelay < 0:
            # Node should have been awake by now. Wait 2 minutes to allow
            # the node to be manually woken
            expectedDelay = 120
        # Add some "wiggle room" for the wakeup
        self._expectedDelay = expectedDelay + 60
        self._status = Ack.Status.QUEUED
        self._event.set()

    async def wait(self, timeout):
        """Wait until the node ack the message"""
        await asyncio.wait_for(self._event.wait(), timeout)
        # Message was queued for a sleeping node.
        # Wait longer!
        while self._status == Ack.Status.QUEUED:
            self._event.clear()
            await asyncio.wait_for(self._event.wait(), self._expectedDelay)


class Adapter(Listenable, MessageWaiter, metaclass=abc.ABCMeta):
    """Abstract class for implementing communication with a Z-Wave chip"""

    def __init__(self):
        super().__init__()
        self._ackQueue = {}
        self._nodeId = 0

    def ackReceived(self, zipPkt: Zip.ZipPacket):
        """Call this method when an ack message has been received"""
        ackId = zipPkt.seqNo
        if zipPkt.nackResponse and zipPkt.nackWaiting:
            # Message was queued, signal this and keep ack in queue
            ack = self._ackQueue.get(ackId, None)
            if not ack:
                return False
            ack.queued(zipPkt.headerExtension.expectedDelay)
            return True
        ack = self._ackQueue.pop(ackId, None)
        if not ack:
            _LOGGER.warning("Received ack %d for command not waiting for", ackId)
            return False
        ack.received()
        return True

    @abc.abstractmethod
    async def addNode(self, txOptions: TxOptions) -> bool:
        """Start inclusion mode in the controller"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def addNodeDSKSet(
        self, accept: bool, inputDSKLength: int, dsk: dsk_t
    ) -> bool:
        """
        This command is used to indicate the S2 bootstrapping controller if the DSK is
        accepted and report the user input when needed.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def addNodeKeysSet(
        self, grantCSA: bool, accept: bool, grantedKeys: NetworkManagementInclusion.Keys
    ) -> bool:
        """
        This command is used to inform an S2 bootstrapping controller which keys must be
        granted to the node being bootstrapped.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def addNodeStop(self) -> bool:
        """Stop inclusion mode in the controller"""
        raise NotImplementedError()

    def commandReceived(self, cmd: Message):
        """Call this method when a command has been received"""
        if isinstance(cmd, Zip.ZipPacket):
            msg = cmd.command
        else:
            msg = cmd
        if not self.messageReceived(msg):
            self.speak("onMessageReceived", cmd)

    @abc.abstractmethod
    async def connect(self):
        """Connect the adapter. Must be implemented by subclass"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def getFailedNodeList(self) -> list:
        """Return a list of failing nodes"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def getMultiChannelCapability(
        self, nodeId: int, endpoint: int
    ) -> NetworkManagementProxy.MultiChannelCapabilityReport:
        """Return the multi channel capabilities for an endpoint in a node"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def getMultiChannelEndPoints(self, nodeId: int) -> int:
        """Return the number of multi channel end points implemented by a node"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def getNodeInfo(
        self, nodeId: int
    ) -> NetworkManagementProxy.NodeInfoCachedReport:
        """Return the node info from this node. Possibly cached"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def getNodeList(self) -> set:
        """Return a list of nodes included in the network"""
        raise NotImplementedError()

    @property
    def nodeId(self) -> int:
        """Return the node id of the controller"""
        return self._nodeId

    @nodeId.setter
    def nodeId(self, nodeId: int):
        self._nodeId = nodeId

    @abc.abstractmethod
    async def removeFailedNode(
        self, nodeId: int
    ) -> NetworkManagementInclusion.FailedNodeRemoveStatus.Status:
        """Remove a non-responding node"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def removeNode(self) -> bool:
        """Start exclusion mode in the controller"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def removeNodeStop(self) -> bool:
        """Stop exclusion mode in the controller"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def send(
        self, cmd: Message, sourceEP: int = 0, destEP: int = 0, timeout: int = 3
    ) -> bool:
        """
        Send message to Z-Wave chip. Must be implemented in subclass.

        .. warning::
          This command will block until the message has been ACKed by the node.

          When talking to battery operated (sleeping) nodes this command will
          block until the nodes wakes up or is considered dead. This can be a
          long time (week or even months). Please make sure the code can handle this.
        """
        raise NotImplementedError()

    async def sendToNode(self, nodeId: int, cmd: Message, **kwargs) -> bool:
        """Send message to node. Must be implemented in subclass"""
        raise NotImplementedError()

    async def sendAndReceive(
        self, cmd: Message, waitFor: Message, timeout: int = 3, **kwargs
    ) -> Message:
        """Send a message and wait for the response"""
        session = self.addWaitingSession(waitFor)
        await self.send(cmd, **kwargs)
        return await self.waitForMessage(waitFor, session=session, timeout=timeout)

    @abc.abstractmethod
    async def setNodeInfo(self, generic, specific, cmdClasses):
        """
        Set the application NIF (Node Information Frame). This method
        should not be called directly. Use the corresponding function
        in Application instead.
        """
        raise NotImplementedError()

    async def waitForAck(self, ackId: int, timeout: int = 3):
        """Async method for waiting for the adapter to receive a specific ack id"""
        if ackId in self._ackQueue:
            raise Exception("Duplicate ackid used!")
        ack = Ack()
        self._ackQueue[ackId] = ack
        try:
            await ack.wait(timeout)
        except asyncio.TimeoutError:
            _LOGGER.warning("Timeout waiting for response for ack %s", ackId)
            del self._ackQueue[ackId]
            raise



# -*- coding: utf-8 -*-

MAJOR_VERSION = 0
MINOR_VERSION = 2
PATCH_VERSION = 0
__short_version__ = "{}.{}".format(MAJOR_VERSION, MINOR_VERSION)
__version__ = "{}.{}".format(__short_version__, PATCH_VERSION)



# -*- coding: utf-8 -*-

import asyncio
import inspect
import logging
from typing import Dict, Any

from pyzwave.types import BitStreamReader

_LOGGER = logging.getLogger(__name__)


class AttributesMixin:
    """Inheritable class to implement defined attributes"""

    attributes = ()

    def __init__(self, **kwargs):
        super().__init__()
        self._attributes = {}
        for attr in getattr(self, "attributes"):
            attrName, attrType = attr[0], attr[1]
            if attrName not in kwargs:
                continue
            value = kwargs[attrName]
            if isinstance(value, AttributesMixin):
                # Is a subclass of ourself, not not wrap it
                self._attributes[attrName] = value
            elif hasattr(attrType, "__setstate__"):
                self._attributes[attrName] = attrType()
                self._attributes[attrName].__setstate__(value)
            else:
                self._attributes[attrName] = attrType(value)

    def attributeUpdated(self, name, newValue, oldValue):
        """Called if an attribute value was updated"""

    def debugString(self, indent=0):
        """
        Convert all attributes in this object to a human readable string used for debug output.
        """
        attrs = []
        for attr in self.attributes:
            attrName = attr[0]
            if attrName not in self._attributes:
                continue
            if hasattr(self._attributes[attrName], "debugString"):
                value = self._attributes[attrName].debugString(indent + 1)
            else:
                value = repr(self._attributes[attrName])
            attrs.append("{}{} = {}".format("\t" * (indent + 1), attrName, value))
        return "{}:\n{}".format(str(self), "\n".join(attrs))

    def parseAttributes(self, stream: BitStreamReader):
        """Populate the attributes from a raw bitstream."""
        for attr in self.attributes:
            if stream.bytesLeft() == 0:
                # No more data, cannot decode rest of the attributes
                break
            attrName, attrType = attr[0], attr[1]
            deserializer = getattr(self, "parse_{}".format(attrName), None)
            if deserializer:
                value = deserializer(stream)
            else:
                value = attrType.deserialize(stream)
            # This can be optimized to reduce the second loop in __setattr__
            setattr(self, attrName, value)

    def __getattr__(self, name):
        if name not in self._attributes:
            # Try to load default
            for attrName, attrType in getattr(self, "attributes"):
                if attrName != name:
                    continue
                attr = attrType()
                self._attributes[name] = attr
                return attr
        return self._attributes.get(name)

    def __getstate__(self) -> Dict[str, Any]:
        values = {}
        for attr, value in self._attributes.items():
            if hasattr(value, "__getstate__"):
                values[attr] = value.__getstate__()
            else:
                values[attr] = value
        return values

    def __setattr__(self, name, value):
        for attr in getattr(self, "attributes"):
            attrName, attrType = attr[0], attr[1]
            if attrName == name:
                oldValue = self._attributes.get(name, None)
                if isinstance(value, attrType):
                    # Correct type set, use it directly
                    newValue = value
                elif hasattr(attrType, "__setstate__"):
                    newValue = attrType()
                    newValue.__setstate__(value)
                elif isinstance(value, tuple):
                    newValue = attrType(*value)
                else:
                    newValue = attrType(value)
                if oldValue == newValue:
                    return
                self._attributes[name] = newValue
                self.attributeUpdated(name, newValue, oldValue)
                return
        super().__setattr__(name, value)

    def __setstate__(self, state):
        for attrName, attrType in getattr(self, "attributes"):
            if attrName not in state:
                continue
            if hasattr(attrType, "__setstate__"):
                value = attrType()
                value.__setstate__(state[attrName])
            else:
                value = attrType(state[attrName])
            self._attributes[attrName] = value


class Listenable:
    """Inheritable class to implement listaner interface between classes"""

    def __init__(self):
        super().__init__()
        self._listeners = []

    def addListener(self, listener):
        """Add class as listener for messages"""
        self._listeners.append(listener)

    async def ask(self, message, *args) -> list:
        """
        Send message to listeners and wait for the listeners to respond.
        This a shorthand for awaiting thre result from speak()
        """
        return await asyncio.gather(*self.speak(message, *args))

    def speak(self, message, *args) -> list:
        """
        Send message to listeners.
        Returns a list of futures if the listeners are async. This can be used to allow waiting
        for all listeners to finish before continue.
        """
        retval = []
        for listener in self._listeners:
            method = getattr(listener, message, None)
            if not method:
                continue
            if inspect.iscoroutinefunction(method):
                # Listener is a coroutine, call async
                retval.append(asyncio.ensure_future(method(self, *args)))
                continue
            try:
                future = asyncio.get_event_loop().create_future()
                future.set_result(method(self, *args))
                retval.append(future)
            except Exception as error:
                _LOGGER.warning("Error calling listener.%s: %s", message, error)
        return retval


class MessageWaiter:
    """Inheritable class to implement listening for specific messages"""

    def __init__(self):
        super().__init__()
        self._sessions = {}

    def addWaitingSession(self, msgType):
        """
        Setup the session to wait for _before_ doing the wait. Do this to avoid
        a race condition where the message is received before we wait for it
        """
        hid = msgType.hid()
        session = self._sessions.get(hid, None)
        if not session:
            session = asyncio.get_event_loop().create_future()
            self._sessions[hid] = session
        return session

    def messageReceived(self, message) -> bool:
        """Called when a message is received directed to this node"""
        session = self._sessions.pop(message.hid(), None)
        if session:
            session.set_result(message)
            return True
        return False

    async def waitForMessage(self, msgType, timeout: int = 3, session=None):
        """Async method for waiting for a specific message to arrive from the node."""
        if not session:
            session = self.addWaitingSession(msgType)
        try:
            await asyncio.wait_for(session, timeout)
        except asyncio.TimeoutError:
            _LOGGER.warning("Timeout waiting for message %s", msgType)
            raise
        finally:
            self._sessions.pop(msgType.hid(), None)
        return session.result()



# -*- coding: utf-8 -*-

import asyncio
import logging
import ssl
import socket
import threading

from ctypes import (
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_void_p,
    CFUNCTYPE,
    POINTER,
    memmove,
)

import dtls


dtls.do_patch()

_LOGGER = logging.getLogger(__name__)

SSL_CB_ALERT = 0x4000  # /* used in callback */

CLIENTPSKFUNC = CFUNCTYPE(
    c_uint, c_void_p, c_char_p, POINTER(c_char), c_uint, POINTER(c_char), c_uint,
)
SERVERPSKFUNC = CFUNCTYPE(c_uint, c_void_p, c_char_p, POINTER(c_char), c_uint)
CLIENTCBFUNC = CFUNCTYPE(None, c_void_p, c_int, c_int)


class DTLSConnection(threading.Thread):
    """Connection object to create a DTLS connection using PSK"""

    def __init__(self):
        super().__init__(name="Z-Wave DTLS connection")
        self._address = None
        self._psk = None
        self._running = False
        self._sock = None
        self._connectionEvent = asyncio.Event()
        self._loop = asyncio.get_event_loop()
        self._onMessage = None
        self._server = False
        self.setDaemon(True)

        self.clientPskCbWrapper = CLIENTPSKFUNC(self.clientPskCb)
        self.clientCbWrapper = CLIENTCBFUNC(self.clientCb)
        self.serverPskCbWrapper = SERVERPSKFUNC(self.serverPskCb)

    async def connect(self, address, psk):
        """Connect to remote using psk"""
        self._address = address
        self._psk = psk
        self._connectionEvent.clear()
        self.start()
        await asyncio.wait_for(self._connectionEvent.wait(), timeout=10)

    def listen(self, psk):
        """Start server socket"""
        self._server = True
        self._psk = psk
        self.start()

    def run(self):  # pylint: disable=missing-function-docstring
        self._sock = self.createDtlsPskSock()
        self._loop.call_soon_threadsafe(self._connectionEvent.set)
        self._running = True
        while self._running:
            try:
                if self._server:
                    pkt, address = self._sock.recvfrom(1500)
                    if self._onMessage:
                        self._loop.call_soon_threadsafe(self._onMessage, pkt, address)
                else:
                    pkt = self._sock.recv(1500)
                    if self._onMessage:
                        self._loop.call_soon_threadsafe(self._onMessage, pkt)
            except AttributeError:
                _LOGGER.error("Got attribute error!")
            except ssl.SSLError as error:
                _LOGGER.info("SSL error: %s", error)
                break
            except Exception as error:
                _LOGGER.error("Could not read from zipgateway %s", error)
                break

    def send(self, msg):
        """Send bytes to socket"""
        if not self._sock:
            _LOGGER.error("Could not send, not yet connected!")
            return
        self._sock.send(msg)

    def onMessage(self, cbfn):
        """Set the callback function to use when data has arrived"""
        self._onMessage = cbfn

    def stop(self):
        """Stop the thread"""
        self._running = False

    def clientPskCb(self, _ssl, _hint, identity, _maxIdenityLen, cpsk, _maxPskLen):
        """Callback function used by ssl to get the DTLS psk"""
        length = len(self._psk)
        memmove(cpsk, self._psk, length)
        iden = "Client_identity"
        memmove(identity, iden.encode("utf-8"), len(iden))
        return length

    def serverPskCb(self, _ssl, _identity, cpsk, _maxPskLen):
        """Callback function used by ssl to get the DTLS psk"""
        length = len(self._psk)
        memmove(cpsk, self._psk, length)
        return length

    def clientCb(self, _ssl, where, ret):  # pylint: disable=missing-function-docstring

        # pylint: disable=protected-access
        if where & SSL_CB_ALERT and self._sock._connected and not self._server:

            print("Client Alert", where, ret, self._sock._connected)
            # sock.unwrap();
            self._sock._connected = True

            self._sock._sslobj = _ssl.sslwrap(  # pylint: disable=no-member
                self._sock._sock,
                False,
                None,
                None,
                _ssl.CERT_NONE,
                _ssl.PROTOCOL_SSLv23,
                None,
                None,
            )

            self._sock.do_handshake()
            # try:
            # 	dtls.openssl.SSL_shutdown(sock._sslobj._ssl.value);
            # except:
            # 	pass
            # sock.shutdown(socket.SHUT_WR)

    def createDtlsPskSock(self):
        """Create a new socket and configure it for DTLS PSK"""
        if self._server:
            sock = dtls.wrap_server(
                socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
                certfile="/tmp/ZIPR.x509_1024.pem",
                keyfile="/tmp/ZIPR.key_1024.pem",
                ca_certs="/tmp/Portal.ca_x509.pem",
                cert_reqs=ssl.CERT_NONE,
                do_handshake_on_connect=False,
                ciphers="PSK",
            )
            sock.bind(("::", 4123,))
            sock.listen(5)
            # sock.settimeout(5.0)

            proto = (
                "SSL_set_psk_server_callback",
                dtls.openssl.libssl,
                ((None, "ret"), (dtls.openssl.SSL, "ctx"), (c_void_p, "psk_server_cb")),
            )
            dtls.openssl._make_function(*proto)  # pylint: disable=protected-access
            # pylint: disable=no-member,protected-access
            dtls.openssl.SSL_set_psk_server_callback(
                sock._sslobj._ssl.value, self.serverPskCbWrapper
            )
        else:
            sock = ssl.wrap_socket(
                socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
                do_handshake_on_connect=False,
                ciphers="PSK",
            )
            # sock.setblocking(True)
            # sock.settimeout(20.0)
            sock.connect((self._address, 41230))

            # Using ctypes we interface with the function SSL_set_psk_client_callback from OpenSSL
            proto = (
                "SSL_set_psk_client_callback",
                dtls.openssl.libssl,
                ((None, "ret"), (dtls.openssl.SSL, "ctx"), (c_void_p, "psk_client_cb")),
            )
            dtls.openssl._make_function(*proto)  # pylint: disable=protected-access
            # pylint: disable=no-member,protected-access
            dtls.openssl.SSL_set_psk_client_callback(
                sock._sslobj._ssl.value, self.clientPskCbWrapper
            )

        proto = (
            "SSL_set_info_callback",
            dtls.openssl.libssl,
            ((None, "ret"), (dtls.openssl.SSL, "ctx"), (c_void_p, "callback")),
        )
        dtls.openssl._make_function(*proto)  # pylint: disable=protected-access

        # pylint: disable=no-member,protected-access
        dtls.openssl.SSL_set_info_callback(
            sock._sslobj._ssl.value, self.clientCbWrapper
        )

        if not self._server:
            sock.do_handshake()
        return sock



"""Setup module for pyzwave"""

from setuptools import setup
import pyzwave

with open("README.md", "r") as fh:
    longDescription = fh.read()  # pylint: disable=invalid-name

setup(
    name="python-zwave",
    version=pyzwave.__version__,
    description="Library implementing a Z-Wave stack",
    long_description=longDescription,
    long_description_content_type="text/markdown",
    url="http://github.com/pyzwave/pyzwave",
    author="Micke Prag",
    author_email="micke.prag@telldus.se",
    license="GPL-3.0",
    packages=["pyzwave"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Home Automation",
    ],
    python_requires=">=3.6",
    project_urls={
        "Documentation": "https://pyzwave.readthedocs.io",
        "Source": "http://github.com/pyzwave/pyzwave",
    },
)
